"""transform a parquet collection into an ordered file"""

from tqdm import tqdm
from struct import pack
import pyarrow as pa
import pyarrow.dataset as ds
import pandas as pd
import fire


def parquet_to_file(
    input_collection,
    output_file,
    key_column="hash",
    value_columns=None,
    key_format="q",
    value_format="ee",
):
    """transform a parquet collection into an ordered file"""
    if value_columns is None:
        value_columns = ["pwatermark", "punsafe"]

    dataset = ds.dataset(input_collection, format="parquet")

    def encodeb(k, values):
        k, v = pack(key_format, k), pack(value_format, *values)
        return k + v

    with open(output_file, "wb") as f:
        for batch in tqdm(dataset.to_batches()):
            d = batch.to_pandas(types_mapper={pa.int64(): pd.Int64Dtype()}.get)
            for t in zip(d[key_column], *[d[value_column] for value_column in value_columns]):
                k = t[0]
                vs = t[1:]
                l = encodeb(k, vs)
                f.write(l)


if __name__ == "__main__":
    fire.Fire(parquet_to_file)



"""order parquet collection"""


import fire


def sort_parquet(input_collection, key_column, value_columns, output_folder):
    """order parquet collection"""
    from pyspark.sql import SparkSession  # pylint: disable=import-outside-toplevel

    spark = SparkSession.getActiveSession()
    if spark is None:
        spark = (
            SparkSession.builder.config("spark.sql.autoBroadcastJoinThreshold", -1)
            .config("spark.sql.shuffle.partitions", 1000)
            .config("spark.driver.memory", "16G")
            .master("local[16]")
            .appName("spark-stats")
            .getOrCreate()
        )
    df = spark.read.parquet(input_collection)
    df = df.select(key_column, *value_columns)
    df = df.filter(df[key_column].isNotNull())
    for v in value_columns:
        df = df.filter(df[v].isNotNull())
    df = df.sort(key_column)
    df.coalesce(512).write.mode("overwrite").parquet(output_folder)


if __name__ == "__main__":
    fire.Fire(sort_parquet)



"""read"""


import bisect
import os
import mmap
from struct import unpack, Struct


class OnDiskKVByIndex:
    def __init__(self, ondiskkv):
        self.ondiskkv = ondiskkv

    def __getitem__(self, i):
        return self.ondiskkv.get_key(i)


class OnDiskKV:
    """on disk kv provides a fast and simple kv store"""

    def __init__(self, file, key_format="q", value_format="ee"):
        self.key_format = key_format
        self.value_format = value_format
        self.key_size = Struct(key_format).size
        self.value_size = Struct(value_format).size
        self.entry_size = self.key_size + self.value_size
        size = os.path.getsize(file)
        self.length = int(size / self.entry_size)
        self.f = open(file, "rb")  # pylint: disable=consider-using-with
        self.mm = mmap.mmap(self.f.fileno(), 0, prot=mmap.PROT_READ)
        self.filename = file
        self.ondiskkv_by_index = OnDiskKVByIndex(self)

    def __getstate__(self):
        return {
            "filename": self.filename,
            "key_format": self.key_format,
            "value_format": self.value_format,
        }

    def __setstate__(self, state):
        self.__init__(state["filename"], state["key_format"], state["value_format"])

    def get_key(self, i):
        return unpack(self.key_format, self.mm[self.entry_size * i : self.entry_size * i + self.key_size])[0]

    def get_value(self, i):
        return unpack(
            self.value_format, self.mm[self.entry_size * i + self.key_size : self.entry_size * i + self.entry_size]
        )

    def __getitem__(self, k):
        i = bisect.bisect_left(self.ondiskkv_by_index, k, lo=0, hi=self.length)
        k2 = self.get_key(i)
        if k != k2:
            raise ValueError(str(k) + " not found")
        return self.get_value(i)
